/*
** ADOBE SYSTEMS INCORPORATED
** Copyright 2008 Adobe Systems Incorporated. All Rights Reserved.
**
** NOTICE:  Adobe permits you to use, modify, and distribute this file in
** accordance with the terms of the Adobe license agreement accompanying it.
** If you have received this file from a source other than Adobe, then your use,
** modification, or distribution of it requires the prior written permission of Adobe.
*/
#ifndef AS3_H
#define AS3_H

#include <stdlib.h>
/*
** AS3 <=> C bridging API
*/

#ifdef __cplusplus
extern "C" {
#endif

/* Release control to Flash and return to this point on the next timer tick */
void flyield(); 

/* ref counted AS3 value type */
typedef struct _AS3_Val *AS3_Val;

/* just a char * but was malloced (so should be freed) */
typedef char *AS3_Malloced_Str;

#ifndef AS3_NO_C_API

/* all values are ref counted
   and you must release all values
   EXCEPT the params object passed
   to a thunk
*/
void AS3_Acquire(AS3_Val obj);
void AS3_Release(AS3_Val obj);
/* (AS3_Val)NULL is AS3 undefined and doesn't need to be released;
   in some cases, this up-front check will give a performance boost
*/
#ifdef __cplusplus
inline void AS3_ReleaseX(AS3_Val obj)
{
  if(obj != NULL)
    AS3_Release(obj);
}
#else
#define AS3_ReleaseX(OBJ) do { AS3_Val obj = (OBJ); if(obj != NULL) AS3_Release(obj); } while(0)
#endif

/* ns::[prop] */
AS3_Val AS3_NSGet(AS3_Val ns, AS3_Val prop);
AS3_Val AS3_NSGetS(AS3_Val ns, const char *prop);

/* obj[prop] */
AS3_Val AS3_Get(AS3_Val obj, AS3_Val prop);
AS3_Val AS3_GetS(AS3_Val obj, const char *prop);

/* obj[prop] = val */
AS3_Val AS3_Set(AS3_Val obj, AS3_Val prop, AS3_Val val);
AS3_Val AS3_SetS(AS3_Val obj, const char *prop, AS3_Val val);

/* typeof obj */
AS3_Malloced_Str AS3_TypeOf(AS3_Val obj);

/* AS3 value creation */
AS3_Val AS3_String(const char *str);
AS3_Val AS3_StringN(const char *str, int len);
AS3_Val AS3_Int(int n);
AS3_Val AS3_Ptr(void *p);
AS3_Val AS3_Number(double n);
AS3_Val AS3_True();
AS3_Val AS3_False();
AS3_Val AS3_Null();
/* undefined is guaranteed to be (AS3_Val)NULL
   (null is NOT!) and does not need to be acquired
   or released though acquire and releasing are valid
   no-ops 
*/
AS3_Val AS3_Undefined();

/* AS3 value conversion */
AS3_Malloced_Str AS3_StringValue(AS3_Val obj);
int AS3_IntValue(AS3_Val obj);
void *AS3_PtrValue(AS3_Val obj);
double AS3_NumberValue(AS3_Val obj);

/* utility to create an array from a type template
   tt => "type,type,type..." where a type can be
   IntType => int
   PtrType => void *
   DoubleType => double
   StrType => char *
   AS3ValType => AS3_Val
*/
AS3_Val AS3_Array(const char *tt, ...);

/* utility to create an object from a type template
   tt => "prop1:type,prop2:type,prop3:type..." where a type can be
   IntType => int
   PtrType => void *
   DoubleType => double
   StrType => char *
   AS3ValType => AS3_Val
*/
AS3_Val AS3_Object(const char *tt, ...);

/* utility to get a bunch of values from an array using the same
   kind of template as AS3_Array */
void AS3_ArrayValue(AS3_Val array, const char *tt, ...);

/* utility to get a bunch of values from an object using the same
   kind of template as AS3_Object */
void AS3_ObjectValue(AS3_Val object, const char *tt, ...);

/* func.apply(thiz, params:Array) */
AS3_Val AS3_Call(AS3_Val func, AS3_Val thiz, AS3_Val params);

/* thiz[func].apply(thiz, params:Array) */
AS3_Val AS3_CallS(const char *func, AS3_Val thiz, AS3_Val params);

/* func.apply(thiz, AS3_Array(tt, ...)) */
AS3_Val AS3_CallT(AS3_Val func, AS3_Val thiz, const char *tt, ...);

/* thiz[func].apply(thiz, AS3_Array(tt, ...)) */
AS3_Val AS3_CallTS(const char *func, AS3_Val thiz, const char *tt, ...);

/* create a C function pointer that thunks to thiz.func(...)

   int (*myfunc)(const char *s, double n) =
     AS3_Shim(someFunc, someThiz, "IntType, StrType, NumberType", false);

   the function pointer is PERMANENT!
*/
void *AS3_Shim(AS3_Val func, AS3_Val thiz, const char *rt, const char *tt,
  int varargs);

/* new constr(parameters in params:Array) */
AS3_Val AS3_New(AS3_Val constr, AS3_Val params);

typedef AS3_Val (*AS3_ThunkProc)(void *data, AS3_Val params);

/* function(...params):* { return proc(data, params); } */
AS3_Val AS3_Function(void *data, AS3_ThunkProc proc);
AS3_Val AS3_FunctionAsync(void *data, AS3_ThunkProc proc);

AS3_Val AS3_FunctionT(void *data, void *proc, const char *rt, const char *tt,
  int varargs);
AS3_Val AS3_FunctionAsyncT(void *data, void *proc, const char *rt,
  const char *tt, int varargs);

/* val instanceof type */
int AS3_InstanceOf(AS3_Val val, AS3_Val type);

/* get the stage object */
AS3_Val AS3_Stage();

/* create a Proxy object that delegates callProperty, etc. to the
   Function vars in the object in the flash_delegate namespace!
   i.e., delgates to flash_delegate::callProperty, not public callProperty
*/
AS3_Val AS3_Proxy();

/* fetch the ByteArray object representing all of ram */
AS3_Val AS3_Ram();

/* read len bytes from byteArraySrc into the memory pointed to by dst */
int AS3_ByteArray_readBytes(void *dst, AS3_Val byteArraySrc, int len);

/* write len bytes from memory pointed to by src into byteArrayDst */
int AS3_ByteArray_writeBytes(AS3_Val byteArrayDst, void *src, int len);

/* adjust the position in the ByteArray according to "lseek" style rules */
int AS3_ByteArray_seek(AS3_Val byteArrayDst, int offs, int whence);

/* [un]registers a jmp_buf as being "abused" -- used for coop-MT or
** conservative GC */
/* when setjmp is called on a registered "abused" jmp_buf, the setjmp
** will call the provided alloc and/or free routines to allocate
** space to store the extra data needed to support "abuse" of setjmp/
** longjmp (provide routines than manipulated GC-ed space in a conservative
** GC environment like, say, Ruby)
*/
typedef void *(*AS3_AllocProc)(int);
typedef void (*AS3_FreeProc)(void *);
void AS3_Reg_jmp_buf_AbuseHelpers(AS3_AllocProc alloc, AS3_FreeProc free);
void AS3_RegAbused_jmp_buf(void *buf);
void AS3_UnregAbused_jmp_buf(void *buf);

/* trace the AS3 val */
void AS3_Trace(AS3_Val val);

/* notify that the library has initialized */
/* (does not return) */
void AS3_LibInit(AS3_Val libData);

#endif /* AS3_NO_C_API */

#ifdef __cplusplus
}

#ifndef AS3_NO_CXX_API
namespace AS3
{
  class Value
  {
    AS3_Val m_val;

  public:
    Value() : m_val(AS3_Undefined()) {}
#ifndef SWIG
    Value(AS3_Val val) : m_val(val) {}
    Value(const Value &v) : m_val(v.m_val) { AS3_Acquire(m_val); }
#endif
    
    ~Value() { AS3_Release(m_val); }

    AS3_Val Get() const { return m_val; }

    void Set(AS3_Val val)
    {
      AS3_Release(m_val);
      m_val = val;
    }

#ifndef SWIG
    operator AS3_Val() const { return Get(); }

    Value &operator=(const Value &v)
    {
      AS3_Acquire(v.m_val);
      Set(v.m_val);
      return *this;
    }

    Value &operator=(AS3_Val val)
    {
      Set(val);
      return *this;
    }
#endif
  };

#ifndef SWIG
  class MStr
  {
    char *m_sz;

    MStr &operator=(const MStr &);
    MStr(const MStr &);

  public:
    MStr() : m_sz(NULL) {}
    MStr(char *sz) : m_sz(sz) {}
    ~MStr() { if(m_sz) free(m_sz); }

    MStr &operator=(char *sz)
      { if(m_sz) free(m_sz); m_sz = sz; return *this; }
    operator char *() const { return m_sz; }
  };
#endif
}
#endif

#endif

#endif /* AS3_H */
